define(['underscore', 'cookiesjs', 'squel'], function (_, Cookies, squel) {
  return {
    db: null,
    dbs: [],
    connection: null,
    table: null,
    queryMode: "filter", // @deprecated
    query_mode: "filter",
    output_mode: "table",
    meta: {},
    section: "content",
    driver: null,
    loaded: false,
    output: true,
    query: null,
    charsets: [],
    lastRawQuery: null,
    result: null,
    quoteChar: function () {
      switch (this.driver) {
        case "redshift": // same as postgres
        case "postgresql":
          return '"';
          break;
        default:
          return '`';
          break;
      }
    },
    quote: function (field) {
      return this.quoteChar() + field + this.quoteChar();
    },
    serializeObj: function ($el) {
      var hash = {};
      return _.reduce($el.serializeArray(), function (hash, pair) {
        hash[pair.name] = pair.value;
        return hash;
      }, {});
    },
    squelSettings: {
      postgresql: {
        autoQuoteTableNames: true,
        autoQuoteFieldNames: false,
        numberedParameters: false
      },
      mysql: {
        autoQuoteTableNames: true,
        autoQuoteFieldNames: false,
        numberedParameters: false
      },
      redshift: {
        autoQuoteTableNames: true,
        autoQuoteFieldNames: false,
        numberedParameters: false
      }
    },
    getSquelSettings: function () {
      if (!_.has(this.squelSettings, this.driver)) {

        // issue a warning if settings not found in object
        console.warn("Driver specific settings not found in configuration, falling back to mysql.",
            this.driver);

        return this.squelSettings["mysql"];
      }

      this.squelSettings[this.driver].nameQuoteCharacter = this.quoteChar();
      return this.squelSettings[this.driver];
    },
    isAdmin: function () {
      return this.user && this.user.get("is_root");
    },
    isMySQL: function () {
      return this.driver === "mysql";
    },
    isPostgres: function () {
      return this.driver === "postgresql";
    },
    isRedshift: function () {
      return this.driver === "redshift";
    },
    getLastRawQuery: function () {
      if (!this.lastRawQuery) {
        this.lastRawQuery = Cookies.get("last_raw_query");
      }
      return this.lastRawQuery;
    },
    setLastRawQuery: function (query) {
      Cookies.set("last_raw_query", query);
      this.lastRawQuery = query;
    },
    getDefaultQuery: function (table) {

      table = table || this.table;

      if (_.isEmpty(table)) {
        // bad
        return "";
      }

      // query builder instance
      var s = squel.select(this.getSquelSettings());

      // set from table name
      s.from(table);

      // set base limit
      s.limit(20);

      return s;
    },
    queryMatchesCurrentTableSchema: function () {
      // check if columns returned by last query belong to current table schema

      // check if there meta is available for current table
      if (_.isEmpty(this.meta) || _.isEmpty(this.meta.columns)) {
        return false;
      }

      // check if meta is available for last query
      if (_.isEmpty(this.result) || _.isEmpty(this.result.meta)) {
        return false;
      }

      // columns of current table schema
      var currentTableColumns = this.meta.columns;
      var lastQueryColumns = _.pluck(this.result.meta, "name");
      var nonSchemaColumns = _.difference(lastQueryColumns, currentTableColumns);

      if (nonSchemaColumns.length === 0) {
        return true;
      }

      return false;
    },
    /**
     * Get table name for the last executed query
     * Get table from meta when available, try different strategies otherwise
     * @returns {*}
     */
    getTableFromQuery: function () {

      // default value
      var table = null;
      var matchesCurrentSchema = this.queryMatchesCurrentTableSchema();

      // if table available from context and query matches schema
      if (this.table && matchesCurrentSchema) {
        return this.table;
      }

      if (this.result && this.result.meta && this.result.meta.length > 0) {
        var meta = this.result.meta[0];
        if (meta.table && !_.isEmpty(meta.table)) {
          table = meta.table;
        } else {
          // table not found in meta, can happen with redshift
        }
      }

      // check if the table we just found is not the same that failed integrity test
      if (table === this.table && !matchesCurrentSchema) {
        // the same table that failed integrity check
        // this can't be the correct table for sure
        table = null;
      }

      // check if table is still empty
      // @fixme this step is redundant, if this.table is set
      //if (_.isEmpty(table)) {
      //
      //  // meta not returned by the server or table was missing from meta
      //
      //  // - try to compare current query with tables default
      //  // get last query executed
      //  var lastQuery = this.lastRawQuery;
      //
      //  // get default query from filter view using buildBaseQuery method
      //  // the same method is currently used in raw editor too
      //  var defaultQuery = this.getDefaultQuery();
      //
      //  // table name must be set in context, for this step
      //  if (!_.isEmpty(this.table) && !_.isEmpty(lastQuery) && !_.isEmpty(defaultQuery)) {
      //    lastQuery = _.trim(lastQuery);
      //    defaultQuery = _.trim(defaultQuery);
      //
      //    if (lastQuery === defaultQuery) {
      //      // the query is the same as the default one, table hasn't changed
      //      table = this.table;
      //    }
      //  }
      //
      //}

      return table;
    },
    getColumnMeta: function (column) {
      // check if meta and columnSchema exists
      if (this.meta && this.meta.columnMeta) {
        return _.find(this.meta.columnMeta, function (item) {
          return item.Field && item.Field === column;
        });
      }
      return null;
    },
    isMac: function () {
      return navigator.platform.toUpperCase().indexOf('MAC') >= 0;
    },
    getMetaKey: function () {
      if (this.isMac()) {
        return "cmd";
      }
      return "ctrl";
    },
    titleCase: function (string) {
      if (_.isString(string)) {
        return string.charAt(0).toUpperCase() + string.slice(1);
      }
      return string;
    },
    getMaxUserCount: function () {
      return parseInt(Cookies.get("user_count"));
    }
  };
});
