define(['jquery', 'underscore', 'backbone', 'tpls', 'views/content/query/raw', 'views/content/query/filter', 'events',
      'context', 'models/query', 'views/content/schema/structure', 'baseview', 'views/content/backups'],
    function ($, _, Backbone, Templates, RawView, FilterView, Events, Context, Query, StructureView, BaseView,
              BackupsView) {

      // reference to current scope
      var parent = this;

      return BaseView.extend({
        template: Templates["content/query"],
        initialize: function () {
          _.bindAll(this, "executeQuery", "setQueryData", "activateRawView", "resizePanel", "loadQuery");

          // listen to query:execute event, it will be fired from other views
          //Events.on("query:execute", this.executeQuery);
          this.listenTo(Events, "query:execute", this.executeQuery);
          this.listenTo(Events, "chart:load", this.activateRawView);
          this.listenTo(Events, "table:created", this.activateStructureView);
          this.listenTo(Events, "query:load", this.loadQuery);
          this.listenTo(Events, "table:shortcut", this.tableShortcut);

          // bind resize event to adjust max height of panel body
          $(window).resize(this.resizePanel);
        },
        tableShortcut: function (mode, query) {
          // get query mode button
          var $btn = this.$(".query-mode-nav .btn[data-mode="+mode+"]");

          // activate query mode
          var e = {target: $btn};
          this.activate(e, true);
        },
        loadQuery: function (query) {

          // proceed only if we are not in raw view, because raw view handles query:load by itself
          if (Context.query_mode !== "raw") {
            // check if query model has actual query string
            if (query.get("query")) {
              // set query for raw query editor
              Context.setLastRawQuery(query.get("query"));
            }

            // activate raw view
            this.activateRawView();
          }
        },
        resizePanel: function () {
          var resizeTimeout;
          var staticOffset = 7;
          var that = this;
          clearTimeout(resizeTimeout);
          resizeTimeout = setTimeout(function () {
            var $panelBody = that.$(".panel-body");
            var po = $panelBody.offset();
            var wh = $(window).height();
            var maxHeight = wh - po.top - staticOffset;
            $panelBody.css("max-height", maxHeight);
          }, 50);
        },
        activateRawView: function () {
          // raw query mode button
          var $raw = this.$(".query-mode-nav .btn[data-mode=raw]");

          // activate raw query mode if it is not active already
          if (!$raw.hasClass("active")) {
            var e = {target: $raw};
            this.activate(e, true);
          }
        },
        activateStructureView: function () {
          // structure query mode button
          var $str = this.$(".query-mode-nav .btn[data-mode=structure]");

          // activate raw query mode if it is not active already
          if (!$str.hasClass("active")) {
            var e = {target: $str};
            this.activate(e, true);
          }
        },
        executeQuery: function (cbk) {
          //  || Context.table === null
          var query = this.getQuery();
          if (Context.db === null || _.isEmpty(query)) {
            // warn when trying to execute query, but the context is not fully loaded
            // this will help to identify bugs later on
            console.info("app not yet fully loaded, can't execute query", query, Context.db);
          } else {

            // trigger event indicating that query is really going to be fetched
            Events.trigger("query:fetching");

            // prepare post data
            var data = {q: query, table: Context.table};

            // fetch query data
            var that = this;
            this.query = new Query();
            this.query.fetch({
              data: data, type: "POST", success: function () {
                that.setQueryData(cbk)
              }, error: this.error
            });

            // store query in context
            Context.query = data.q;

            // log raw query and current context
            console.info("query:execute", query, Context);
          }
        },
        error: function (model, error) {
          Events.trigger("error", "query");
          if (error["responseJSON"] && "error" in error["responseJSON"]) {
            alert("Error: %(error)s".format(error["responseJSON"]));
          }
        },
        setQueryData: function (cbk) {
          console.log("set query data", this.query.toJSON());

          if (this.query.get("error")) {
            // if query returned an error, do not trigger query:fetched error, but show an error instead
            alert(this.query.get("error"));
          } else {
            // context is finally finished loading
            Context.loaded = true;
            Context.result = this.query.toJSON();

            // trigger query:fetched event, data output views are listening to this event
            Events.trigger("query:fetched", this.query.toJSON());

            if (cbk) {
              cbk();
            }
          }
        },
        getQuery: function () {
          if (this.subviews["main"] && typeof(this.subviews["main"].getQuery) !== "undefined") {
            // get query from active query mode view
            return this.subviews["main"].getQuery();
          }
          console.log("QueryView.getQuery: No subviews found, can't execute getQuery method");
          return null;
        },
        events: {
          'click .query-mode-nav .btn': 'activate'
        },
        activate: function (e, preventEvents) {

          var views = {
            filter: FilterView,
            raw: RawView,
            structure: StructureView,
            backups: BackupsView
          };

          // remove previous view - prevents duplicate event calling
          if (this.subviews["main"]) {
            this.subviews["main"].remove();
          }

          // set active item class and remove it from other elements
          $(e.target).addClass("active")
              .siblings()
              .removeClass("active");

          // load view
          var mode = $(e.target).data("mode");
          this.subviews["main"] = new views[mode]();

          // update url
          Context.query_mode = mode;
          Backbone.history.syncURL();

          // render query mode view
          this.$(".tab-content").html(this.subviews["main"].render().el);

          // trigger queryMode:change event, when mode is activated
          // this is used to populate data in to template from other views
          Events.trigger("queryMode:change", mode);

          if (!preventEvents && Context.query_mode !== "raw") {
            // trigger query:execute event
            Events.trigger("query:execute");
          }
        },
        render: function () {
          this.$el.html(this.template());

          // @todo refactor look at activateStructureView etc
          this.$(".query-mode-nav .btn").filter(function () {
            return $(this).data("mode") === Context.query_mode;
          }).trigger("click");

          // resize panel to fix viewport
          this.resizePanel();

          return this;
        }
      });
    });
